#!/bin/sh
#
#  palModuleMgr.sh
#  NI-PAL Module Manager Utility for Linux
#  script used to un/install and un/load NI-PAL based kernel drivers and NI-PALSM
#  based user daemon services
#
#  Copyright 2013,
#  National Instruments Corporation.
#  All rights reserved.
#
#  Originated: 05/11/2004
#

# global variables
kernelExtension=.o
nikalKernelInstaller="/usr/local/natinst/nikal/bin/nikalKernelInstaller.sh"
natinstNIPALBootDir=/etc/natinst/nipal/bootdrivers
daemonExtension=.so
userDaemonTool=nipalsm
userDaemonToolPath=/usr/local/sbin
defaultUserDaemonStartRCLevel=99 #at some point an option should be added to set this
defaultUserDaemonKillRCLevel=0 #at some point an option should be added to set this
natinstUserDaemonInstallDir=/etc/natinst/nipal/services
linuxUserDaemonDemandDir=/etc/init.d
natinstNIPALUserDaemonDemandDir=/etc/natinst/nipal/init.d
moduleInstallDirName=""

init_natinstNIPALInstallDir()
{
   natinstNIPALInstallDir="`cat /etc/natinst/nipal/nipal.dir`"
   if [ "" = "$natinstNIPALInstallDir" ]
   then
      echo "NI-PAL is not installed, install and try again" 1>&2
      exit 1
   fi
}

echoUsage()
{
   echo "Usage:" 1>&2
   echo "NI-PAL Module Manager Utility 19.5.0" 1>&2
   echo "" 1>&2
   echo "Usage: palModuleMgr [-i -tTYPE -sSTART][-dDEPENDENCIES]" 1>&2
   echo "                                         [-oOTHER][-f][-c] MODULE |" 1>&2
   echo "                      [-r][-oOTHERS][-c] MODULE |" 1>&2
   echo "                      [-l][-oOTHERS] MODULE |" 1>&2
   echo "                      [-u][-oOTHERS] MODULE |" 1>&2
   echo "option:" 1>&2
   echo "i install" 1>&2
   echo "  t type           - can be \"kernelDriver\" or \"userDaemon\"" 1>&2
   echo "  s start          - can be \"boot\" or \"demand\"" 1>&2
   echo "  d dependencies   - a string of modules depended on" 1>&2
   echo "                     put ',' between modules if neccessary" 1>&2
   echo "  o others         - a string of miscellaneous platform options" 1>&2
   echo "                     <platform1>:<option1>=<value1>," 1>&2
   echo "                                 <option2>=<value2>;" 1>&2
   echo "                     <platform2>:<option3>=<value3>," 1>&2
   echo "                                 ..." 1>&2
   echo "                     ..." 1>&2
   echo "  f force          - force installation" 1>&2
   echo "  c copy           - copy binary from the location" 1>&2
   echo "                     specified as part of MODULE to" 1>&2
   echo "                     the corresponding system directory" 1>&2
   echo "" 1>&2
   echo "r remove" 1>&2
   echo "  o others         - a string of miscellaneous platform options" 1>&2
   echo "                     <platform1>:<option1>=<value1>," 1>&2
   echo "                                 <option2>=<value2>;" 1>&2
   echo "                     <platform2>:<option3>=<value3>," 1>&2
   echo "                                 ..." 1>&2
   echo "                     ..." 1>&2
   echo "  c copy           - delete binary from the corresponding system" 1>&2
   echo "                     directory" 1>&2
   echo "" 1>&2
   echo "l load" 1>&2
   echo "  o others         - a string of miscellaneous platform options" 1>&2
   echo "                     <platform1>:<option1>=<value1>," 1>&2
   echo "                                 <option2>=<value2>;" 1>&2
   echo "                     <platform2>:<option3>=<value3>," 1>&2
   echo "                                 ..." 1>&2
   echo "                     ..." 1>&2
   echo "u unload" 1>&2
   echo "  o others         - a string of miscellaneous platform options" 1>&2
   echo "                     <platform1>:<option1>=<value1>," 1>&2
   echo "                                 <option2>=<value2>;" 1>&2
   echo "                     <platform2>:<option3>=<value3>," 1>&2
   echo "                                 ..." 1>&2
   echo "                     ..." 1>&2
   echo "MODULE             - the name of the module, for example:" 1>&2
   echo "                     - export/targets/kernelk-unversioned.o (for install)" 1>&2
   echo "                     - kernelk (for other operations)" 1>&2
}

# this function depends on natinstUserDaemonInstallDir and moduleNameBase being set
verifyUserModeLibraryExistence()
{
   ls $natinstNIPALUserDaemonDemandDir | grep $moduleNameBase > /dev/null 2>&1
   return $?
}

# this function depends on the variables moduleInstallDirName and moduleNameBase being defined
verifyKernelDriverExistence()
{
   # build up nikalKernelInstaller.sh arguments
   if [ "" = "$moduleInstallDirName" ]
   then
      nikalKernelInstallerArgs="-e ${moduleNameBase}-unversioned${kernelExtension}"
   else
      nikalKernelInstallerArgs="-e -d ${moduleInstallDirName} ${moduleNameBase}-unversioned${kernelExtension}"
   fi

   "$nikalKernelInstaller" $nikalKernelInstallerArgs > /dev/null 2>&1
   return $?
}

parseArguments()
{
   moduleCopy=0
   moduleForce=0

   while [ "" != "$1" ]
   do
      case $(echo "$1" | cut -c -2) in
         -i)moduleOperation=install;;
         -r)moduleOperation=remove;;
         -l)moduleOperation=load;;
         -u)moduleOperation=unload;;
         -c)moduleCopy=1;;
         -f)moduleForce=1;;
         -t)
            if [ "-t" = "$1" ]
            then
               shift
               moduleType=$1
            else
               moduleType=$(echo "$1" | cut -c 3-)
            fi
            if [ "kernelDriver" != $moduleType ] && [ "userDaemon" != $moduleType ]
            then
               echo "unknown module type: $moduleType" 1>&2
               exit 1
            fi
            ;;
         -s)
            if [ "-s" = "$1" ]
            then
               shift
               moduleStart=$1
            else
               moduleStart=$(echo "$1" | cut -c 3-)
            fi
            if [ "boot" != $moduleStart ] && [ "demand" != $moduleStart ]
            then
               echo "unknown module start time: $moduleStart" 1>&2
               exit 1
            fi
            ;;
         -d)
            if [ "-d" = "$1" ]
            then
               shift
               moduleDependency=$1
            else
               moduleDependency=$(echo "$1" | cut -c 3-)
            fi
            ;;
         -o)
            if [ "-o" = "$1" ]
            then
               shift
               moduleOption=$1
            else
               moduleOption=$(echo "$1" | cut -c 3-)
            fi
            if [ "$(echo "$moduleOption" | cut -c -5)" = "linux" ]
            then
               moduleOption=$(echo "$moduleOption" | cut -c 7-)
               if [ "$(echo "$moduleOption" | cut -c -3)" = "dir" ]
               then
                  moduleInstallDirName=$(echo "$moduleOption" | cut -c 5-)
               fi
            fi
            ;;
         *)
            if [ "-" = "$(echo "$1" | cut -c -1)" ]
            then
               echo "Unrecognized Option: $1" 1>&2
               echoUsage
               exit 1
            elif [ "" != "$moduleName" ]
            then
               echo "Received module, $moduleName, already. Cannot accept another module name: $1" 1>&2
               echoUsage
               exit 1
            else
               moduleName=$1
            fi
            ;;
      esac
      shift
   done

   #process module name
   moduleNameBase=${moduleName##*/}    # get rid of directory names
   modulePath=${moduleName%\/$moduleNameBase}
   moduleNameBase=${moduleNameBase#lib}    # get rid of "lib" on user mode stuff
   moduleNameBase=${moduleNameBase%%.*} # remove extension
   moduleNameBase=${moduleNameBase%-unversioned}

   if [ "$modulePath" = "$moduleNameBase" ]
   then
      modulePath=""
   fi

   # post process modulePath to absolute path in case it is a relative path
   if [ "$modulePath" != "" ]
   then
      currentPwd=`pwd`
      cd $modulePath
      modulePath=`pwd`
      cd "$currentPwd"
   fi

   if [ "" = "$moduleName" ] || [ "" = "$moduleNameBase" ]
   then
      echo "module name missing!" 1>&2
      echoUsage
      exit 1
   fi

   if [ "install" = "$moduleOperation" ]
   then
      if [ "" = "$moduleType" ] || [ "" = "$moduleStart" ]
      then
         echo "-i needs -tTYPE and -sSTART options" 1>&2
         echoUsage
         exit 1
      fi
   else
      # determine moduleType for other options (remove, etc...)
      if verifyKernelDriverExistence && verifyUserModeLibraryExistence
      then
         echo "$moduleOperation: ambiguous module: $moduleNameBase" 1>&2
         exit 1
      fi
      if ! verifyKernelDriverExistence && ! verifyUserModeLibraryExistence
      then
         echo "$moduleOperation: no existing module: $moduleNameBase" 1>&2
         exit 1
      fi
      if verifyKernelDriverExistence
      then
         moduleType="kernelDriver"
      else
         moduleType="userDaemon"
      fi
   fi

   # final global name to set up now that we know the exact service type
   if [ "kernelDriver" = "$moduleType" ]
   then
      moduleNameFinal=${moduleNameBase}$kernelExtension
      moduleNameUnversioned=${moduleNameBase}-unversioned${kernelExtension}
   else
      moduleNameFinal=lib${moduleNameBase}$daemonExtension
      if [ "" != $modulePath ]
      then
         moduleNameFinal=`ls $modulePath | grep "${moduleNameFinal}\(\.[0-9]*\)*$" | head -n 1`
      else
         moduleNameFinal=`ls $natinstUserDaemonInstallDir | grep "${moduleNameFinal}\(\.[0-9]*\)*$" | head -n 1`
      fi
   fi
}

### KERNEL DRIVER FUNCTIONS ###

installkernelDriver()
{
   if [ $moduleCopy -eq 0 ]
   then
      echo "$0 -i requires -c in Linux!" 1>&2
      echo "Cannot install the module without copying the file to the system directory" 1>&2
      return 1
   fi

   # remove (and removal will unload if necessary)
   removekernelDriver

   if [ "$moduleStart" = "boot" ]
   then
      echo "$moduleNameBase" > $natinstNIPALBootDir/$moduleNameBase.rc
   fi

   # build up nikalKernelInstaller.sh arguments
   if [ "$moduleInstallDirName" = "" ]
   then
      nikalKernelInstallerArgs="-i $modulePath/$moduleNameUnversioned"
   else
      nikalKernelInstallerArgs="-i -d $moduleInstallDirName $modulePath/$moduleNameUnversioned"
   fi
   # Now do the installation
   if ! "$nikalKernelInstaller" $nikalKernelInstallerArgs
   then
      echo "Failed to execute: "$nikalKernelInstaller" $nikalKernelInstallerArgs" 1>&2
      return 1
   fi

   return 0
}

removekernelDriver()
{
   if [ $moduleCopy -eq 0 ]
   then
      echo "$0 -r requires -c in Linux!" 1>&2
      echo "Cannot uninstall the module without deleting the file from the system directory" 1>&2
      return 1
   fi

   # build up nikalKernelInstaller.sh arguments
   if [ "$moduleInstallDirName" = "" ]
   then
      nikalKernelInstallerArgs="-r $moduleNameUnversioned"
   else
      nikalKernelInstallerArgs="-r -d $moduleInstallDirName $moduleNameUnversioned"
   fi

   if ! "$nikalKernelInstaller" $nikalKernelInstallerArgs >/dev/null 2>&1
   then
      echo "Failed to execute: "$nikalKernelInstaller" $nikalKernelInstallerArgs" 1>&2
   fi

   rm -f $natinstNIPALBootDir/$moduleNameBase.rc
}

### USER DAEMON FUNCTIONS ###

generateUserDaemonScript()
{
   echo "#!/bin/sh" > $1
   echo "#" >> $1
   echo "# Copyright `date "+%C%y"`," >> $1
   echo "# National Instruments Corporation." >> $1
   echo "# All rights reserved." >> $1
   echo "#" >> $1
   echo "# File: $moduleNameBase" >> $1
   echo "#" >> $1
   if [ -x /sbin/chkconfig ]; then
      echo "#  following is the init header information for RH Linux variants, used by" >> $1
      echo "#  /sbin/chkconfig" >> $1
      echo "# chkconfig:         2345 $defaultUserDaemonStartRCLevel $defaultUserDaemonKillRCLevel" >> $1
      echo "# description:       Start National Instruments $moduleNameBase service." >> $1
      echo "#" >> $1
   fi
   if [ -x /sbin/insserv ]; then
      echo "# following is the init header information for SuSE Linux" >> $1
      echo "### BEGIN INIT INFO" >> $1
      echo "# Provides:          $moduleNameBase" >> $1
      echo "# Required-Start:    nipal \$syslog \$remote_fs" >> $1
      echo "# X-UnitedLinux-Should-Start: \$time" >> $1
      echo "# Required-Stop:     nipal \$syslog \$remote_fs" >> $1
      echo "# X-UnitedLinux-Should-Stop: \$time" >> $1
      echo "# Default-Start:     2 3 5" >> $1
      echo "# Default-Stop:      0 1 6" >> $1
      echo "# Short-Description: $moduleNameBase" >> $1
      echo "# Description:       Start National Instruments $moduleNameBase service." >> $1
      echo "### END INIT INFO" >> $1
      echo " " >> $1
      if [ "$moduleStart" = "boot" ]; then
         echo "if [ -x /sbin/insserv ]; then" >> $1
         echo "   ${moduleNameBase}Bin=/usr/sbin/rc$moduleNameBase" >> $1
         echo "   test -x \$${moduleNameBase}Bin || exit 5" >> $1
         echo "fi" >> $1
      fi
   fi
   echo " " >> $1
   echo "isServiceRunning() {" >> $1
   echo "   if [ \"\$(ps -weo pid,cmd | grep $moduleNameBase | grep $userDaemonTool | wc -l 2> /dev/null)\" -gt 0 ]; then" >> $1
   echo "      return 0" >> $1
   echo "   fi" >> $1
   echo "   return 1" >> $1
   echo "}" >> $1
   echo " " >> $1
   echo "signalServiceStop() {" >> $1
   echo "   if ! isServiceRunning; then return 0; fi" >> $1
   echo "   serviceProcessNumber=\`ps -weo pid,cmd | grep $moduleNameBase | head -n 1 | awk '{print \$1 }'\` 2> /dev/null" >> $1
   echo "   echo -n Stopping $moduleNameBase\(\$serviceProcessNumber\)" >> $1
   echo "   kill -INT \$serviceProcessNumber &> /dev/null" >> $1
   echo "   killStatus=\$?" >> $1
   echo "   if [ \$killStatus != 0 ]; then" >> $1
   echo "      echo ." >> $1
   echo "      echo Error sending INT signal to $moduleNameBase." >> $1
   echo "   else" >> $1
   echo "      stopTimeout=0" >> $1
   echo "      timeLimit=10" >> $1
   echo "      while [ \$stopTimeout -lt \$timeLimit ]; do" >> $1
   echo "         if ps axw -o pid | grep -w \$serviceProcessNumber > /dev/null; then" >> $1
   echo "            stopTimeout=\`expr \$stopTimeout + 1\`" >> $1
   echo "            echo -n ." >> $1
   echo "            sleep 1" >> $1
   echo "         else" >> $1
   echo "            echo ." >> $1
   echo "            echo $moduleNameBase stopped." >> $1
   echo "            break" >> $1
   echo "         fi" >> $1
   echo "      done" >> $1
   echo "      if [ \$stopTimeout -eq \$timeLimit ]; then" >> $1
   echo "         echo ." >> $1
   echo "         echo $moduleNameBase did not stop within \$timeLimit seconds, giving up." >> $1
   echo "      fi" >> $1
   echo "   fi" >> $1
   echo "   return \$killStatus" >> $1
   echo "}" >> $1
   echo " " >> $1
   echo "startService() {" >> $1
   echo "   if isServiceRunning; then return 1; fi" >> $1
   echo "   $userDaemonToolPath/$userDaemonTool $natinstUserDaemonInstallDir/$moduleNameFinal &> /dev/null" >> $1
   echo "   if [ -d /var/lock/subsys ]; then" >> $1
   echo "      touch /var/lock/subsys/$moduleNameBase" >> $1
   echo "   fi" >> $1
   echo "   return \$?" >> $1
   echo "}" >> $1
   echo " " >> $1
   echo "stopService() {" >> $1
   echo "   signalServiceStop $moduleNameBase" >> $1
   echo "   if [ -f /var/lock/subsys/$moduleNameBase ]; then" >> $1
   echo "      rm -f /var/lock/subsys/$moduleNameBase" >> $1
   echo "   fi" >> $1
   echo "   return \$?" >> $1
   echo "}" >> $1
   echo " " >> $1
   echo "case \"\$1\" in" >> $1
   echo "   start)        startService ;;" >> $1
   echo "   stop)         stopService ;;" >> $1
   echo "   restart)      stopService" >> $1
   echo "                 startService ;;" >> $1
   echo "   force-reload) stopService" >> $1
   echo "                 startService ;;" >> $1
   echo "   status)       isServiceRunning ;;" >> $1
   echo "   *)            echo \"Usage: \$0 {start|stop|restart|force-reload|status}\"" >> $1
   echo "                 exit 2 ;;" >> $1
   echo "esac" >> $1
   echo "exit \$?" >> $1
   echo " " >> $1
}
cleanupInitScripts()
{
   if [ -e $linuxUserDaemonDemandDir/$moduleNameBase ]; then
      if [ -x /sbin/insserv ]; then
         if [ -e $linuxUserDaemonDemandDir/$moduleNameBase ]; then
            /sbin/insserv -r $linuxUserDaemonDemandDir/$moduleNameBase
         fi
         rm -f /usr/sbin/rc$moduleNameBase
      elif [ -x /sbin/chkconfig ]; then
         /sbin/chkconfig $moduleNameBase --del
      else
         if [ -d /etc/rc0.d ]; then
            for i in 0 1 6; do
               rm -f /etc/rc$i.d/K[0-9][0-9]$moduleNameBase
            done
            for i in $(seq 2 5); do
               rm -f /etc/rc$i.d/S[0-9][0-9]$moduleNameBase
            done
         elif [ -d /etc/init.d/rc0.d ]; then
            for i in 0 1 6; do
               rm -f /etc/init.d/rc$i.d/K[0-9][0-9]$moduleNameBase
            done
            for i in $(seq 2 5); do
               rm -f /etc/init.d/rc$i.d/S[0-9][0-9]$moduleNameBase
            done
         fi
      fi
      rm -f $linuxUserDaemonDemandDir/$moduleNameBase
   fi
}

installuserDaemon()
{
   init_natinstNIPALInstallDir

   # if not installed to correct location already, deal with removing it first then install
   if [ $modulePath != $natinstUserDaemonInstallDir ]
   then
      if ! removeuserDaemon
      then
         if [ $moduleForce -eq 1 ]
         then
            echo "Failed to remove user daemon, but -f is specified, so $0 will install a new module image" 1>&2
            echo "However, you need to reload the service manually" 1>&2
         else
            echo "Failed to remove user daemon. Installation failed!" 1>&2
            echo "Please specify -f to force installation" 1>&2
            return 1
         fi
      fi

      # Now do the installation
      if [ $moduleCopy -eq 1 ]
      then
         if ! cp $modulePath/$moduleNameFinal $natinstUserDaemonInstallDir/$moduleNameFinal
         then
            echo "cp $modulePath/$moduleNameFinal $natinstUserDaemonInstallDir/$moduleNameFinal failed" 1>&2
            return 1
         fi
      else
         # make a soft link
         if ! ln -fs $modulePath/$moduleNameFinal $natinstUserDaemonInstallDir/$moduleNameFinal
         then
            echo "ln -fs $modulePath/$moduleNameFinal $natinstUserDaemonInstallDir/$moduleNameFinal failed" 1>&2
            return 1
         fi
      fi
   else
      if [ ! -e $natinstUserDaemonInstallDir/$moduleNameFinal ]
      then
         echo "Failed to find $moduleNameFinal in $natinstUserDaemonInstallDir" 1>&2
         return 1
      fi
   fi

   generateUserDaemonScript $natinstNIPALUserDaemonDemandDir/$moduleNameBase
   chmod u+x $natinstNIPALUserDaemonDemandDir/$moduleNameBase

   cleanupInitScripts

   if ! ln -fs $natinstNIPALUserDaemonDemandDir/$moduleNameBase $linuxUserDaemonDemandDir/$moduleNameBase
   then
      echo "ln -fs $natinstNIPALUserDaemonDemandDir/$moduleNameBase $linuxUserDaemonDemandDir/$moduleNameBase failed" 1>&2
      return 1
   fi

   # Set up boot time loading, make a soft link to startup stuff
   if [ "$moduleStart" = "boot" ]
   then
      if [ -x /sbin/insserv ]; then
         ln -sf $linuxUserDaemonDemandDir/$moduleNameBase /usr/sbin/rc$moduleNameBase
         /sbin/insserv $linuxUserDaemonDemandDir/$moduleNameBase
      elif [ -x /sbin/chkconfig ]; then
         /sbin/chkconfig $moduleNameBase --add
      else
         if [ -d /etc/rc0.d ]; then
            for i in 0 1 6; do
               ln -sf $linuxUserDaemonDemandDir/$moduleNameBase /etc/rc$i.d/K$defaultUserDaemonKillRCLevel$moduleNameBase
            done
            for i in $(seq 2 5); do
               ln -sf $linuxUserDaemonDemandDir/$moduleNameBase /etc/rc$i.d/S$defaultUserDaemonStartRCLevel$moduleNameBase
            done
         elif [ -d /etc/init.d/rc0.d ]; then
            for i in 0 1 6; do
               ln -sf $linuxUserDaemonDemandDir/$moduleNameBase /etc/init.d/rc$i.d/K$defaultUserDaemonKillRCLevel$moduleNameBase
            done
            for i in $(seq 2 5); do
               ln -sf $linuxUserDaemonDemandDir/$moduleNameBase /etc/init.d/rc$i.d/S$defaultUserDaemonStartRCLevel$moduleNameBase
            done
         else
            echo "rc0.d directory not found, startup script for $moduleNameBase not being registered with init levels."
         fi
      fi
   fi

   return 0
}

removeuserDaemon()
{
   if [ -e $natinstUserDaemonInstallDir/$moduleNameFinal ] && [ "" != $moduleName ]
   then
      if ! unloaduserDaemon
      then
         echo "Failed to unload user daemon, removing the file anyway..." 1>&2
         echo "Please manually unload the user daemon from memory later" 1>&2
      fi

      # remove if it is a soft link to the library
      if [ -L $natinstUserDaemonInstallDir/$moduleNameFinal ]
      then
         rm -f $natinstUserDaemonInstallDir/$moduleNameFinal
      else
         # remove if we are told to do so
         if [ $moduleCopy -eq 1 ] && [ -f $natinstUserDaemonInstallDir/$moduleNameFinal ]
         then
            rm -f $natinstUserDaemonInstallDir/$moduleNameFinal
         fi
      fi
   fi

   cleanupInitScripts

   rm -f $natinstNIPALUserDaemonDemandDir/$moduleNameBase

   return 0
}

loaduserDaemon()
{
   if [ -f $natinstNIPALUserDaemonDemandDir/$moduleNameBase ]
   then
      $natinstNIPALUserDaemonDemandDir/$moduleNameBase start
      return $?
   fi
   echo "Failed to load user daemon: $moduleNameBase. $natinstNIPALUserDaemonDemandDir/$moduleNameBase does not exist" 1>&2
   return 1
}

unloaduserDaemon()
{
   if [ -f $natinstNIPALUserDaemonDemandDir/$moduleNameBase ]
   then
      $natinstNIPALUserDaemonDemandDir/$moduleNameBase stop
      return $?
   fi
   return 0
}

### MAIN ###

# check root access
if [ `id -ur` != 0 ]; then
   echo "error: $0 must be run as root" 1>&2
   exit 1
fi

parseArguments $*

#call operation command
$moduleOperation$moduleType
